import os
from functools import cached_property

from utils import Log

log = Log("AbstractDocRemotePathMixin")


class AbstractDocRemotePathMixin:

    @classmethod
    def get_github_username(cls) -> str:
        assert os.environ["GITHUB_USERNAME"]
        return os.environ["GITHUB_USERNAME"]

    @classmethod
    def get_remote_repo_url(cls) -> str:
        # E.g. https://github.com/nuuuwan/lk_acts
        return "/".join(
            [
                "https://github.com",
                cls.get_github_username(),
                cls.get_repo_name(),
            ]
        )

    @classmethod
    def get_data_branch_name(cls):
        if not cls.is_multi_doc():
            return "data"
        return f"data_{cls.get_doc_class_label()}"

    @classmethod
    def get_remote_data_url_base(cls) -> str:
        # E.g. https://github.com/nuuuwan/lk_acts/tree/data
        return "/".join(
            [
                cls.get_remote_repo_url(),
                "tree",
                cls.get_data_branch_name(),
            ]
        )

    @classmethod
    def get_remote_data_url_for_class(cls) -> str:
        # E.g. https://github.com/nuuuwan/lk_acts/tree/data/lk_acts
        return "/".join(
            [
                cls.get_remote_data_url_base(),
                cls.get_dir_docs_for_cls_relative(),
            ]
        )

    @cached_property
    def remote_data_url(self) -> str:
        return "/".join(
            [
                self.__class__.get_remote_data_url_for_class(),
                self.dir_doc_relative_to_class,
            ]
        )

    @classmethod
    def get_raw_remote_data_branch_url(cls) -> str:
        # E.g. https://raw.githubusercontent.com/nuuuwan/lk_appeal_court_judgements/refs/heads/data # noqa: E501
        return "/".join(
            [
                "https://raw.githubusercontent.com",
                cls.get_github_username(),
                cls.get_repo_name(),
                "refs",
                "heads",
                cls.get_data_branch_name(),
            ]
        )



import inspect
import os
import pathlib
from dataclasses import asdict
from functools import cached_property

from utils import FileOrDirectory, JSONFile, Log, TSVFile

log = Log("AbstractDocMetadataMixin")


class AbstractDocMetadataMixin:
    MAX_X_LABELS = 5

    def to_dict(self) -> dict:
        return dict(
            doc_type=self.get_doc_class_label(),
            doc_id=self.doc_id,
        ) | asdict(self)

    def write(self, force: bool = False):
        if not force and self.has_metadata:
            return
        os.makedirs(self.dir_doc, exist_ok=True)
        JSONFile(self.json_path).write(self.to_dict())
        log.info(f"Wrote {self.json_path}")

    @classmethod
    def get_all_json_paths(cls) -> list[str]:
        return [
            str(json_path)
            for json_path in pathlib.Path(cls.get_dir_docs_for_cls()).rglob(
                "doc.json"
            )
        ]

    @classmethod
    def from_dict(cls, d: dict):
        sig = inspect.signature(cls.__init__)
        valid_keys = set(sig.parameters) - {"self"}
        filtered_data = {k: v for k, v in d.items() if k in valid_keys}
        return cls(**filtered_data)

    @classmethod
    def from_file(cls, json_path: str):
        d = JSONFile(json_path).read()
        return cls.from_dict(d)

    @cached_property
    def cmp(self):
        return (self.date_str, self.doc_id)

    @classmethod
    def list_all(cls):
        doc_list = [
            cls.from_file(json_path) for json_path in cls.get_all_json_paths()
        ]
        doc_list.sort(key=lambda doc: doc.cmp, reverse=True)
        return doc_list

    @classmethod
    def list_all_ascending(cls):
        doc_list = cls.list_all()
        doc_list.sort(key=lambda doc: doc.cmp, reverse=False)
        return doc_list

    @classmethod
    def get_url_metadata_set(cls) -> set[str]:
        return {
            doc.url_metadata
            for doc in cls.list_all()
            if doc.url_metadata is not None
        }

    @classmethod
    def get_best_time_unit(cls) -> str:
        doc_list = cls.list_all()
        for time_unit in ["decade", "year", "month", "day"]:
            ts_list = [doc.get_ts(time_unit) for doc in doc_list]
            ts_to_n = {}
            for ts in ts_list:
                ts_to_n[ts] = ts_to_n.get(ts, 0) + 1
            n = len(ts_list)
            min_n = n * 0.01
            sufficient_ts_list = [
                x[0] for x in ts_to_n.items() if x[1] >= min_n
            ]
            n_sufficient_ts = len(set(sufficient_ts_list))
            if n_sufficient_ts >= cls.MAX_X_LABELS:
                return time_unit
        return "year"

    def get_ts(self, time_unit: str) -> str:
        ts = {
            "decade": self.date_str[:3] + "0s",
            "year": self.date_str[:4],
            "month": self.date_str[:7],
            "day": self.date_str[:10],
        }.get(time_unit, None)
        if ts is None:
            raise ValueError(f"Invalid time_unit: {time_unit}")
        return ts

    @classmethod
    def get_ts_to_lang_to_n(cls, time_unit: str):
        idx = {}
        for doc in cls.list_all():
            ts = doc.get_ts(time_unit)
            lang = doc.lang
            if ts not in idx:
                idx[ts] = {}
            if lang not in idx[ts]:
                idx[ts][lang] = 0
            idx[ts][lang] += 1
        return idx

    @classmethod
    def get_all_tsv_path(cls, suffix) -> str:
        # E.g. "../lk_acts_data/data/lk_acts/all.tsv"
        return os.path.join(cls.get_dir_docs_for_cls(), f"docs_{suffix}.tsv")

    @classmethod
    def write_all(cls):
        d_list = [doc.to_dict() for doc in cls.list_all()]

        for n, suffix in [
            [None, "all"],
            [100, "last100"],
            [1000, "last1000"],
            [10000, "last10000"],
        ]:
            if n and n > len(d_list):
                continue

            d_list_for_file = d_list[:n] if n else d_list
            all_tsv_path = cls.get_all_tsv_path(suffix)
            TSVFile(all_tsv_path).write(d_list_for_file)
            log.info(f"Wrote {FileOrDirectory(all_tsv_path)}")



import os

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.ticker import MaxNLocator
from utils import Log

log = Log("AbstractDocChartDocsByTimeAndLangMixin")


class AbstractDocChartDocsByTimeAndLangMixin:
    FIG_SIZE = (12, 6.75)
    LANGS = ["si", "ta", "en"]
    COLOR_MAP = {
        "si": "#8D153A",
        "ta": "#EB7400",
        "en": "#00534E",
        "si-ta-en": "#FFBE29",
    }

    @classmethod
    def get_chart_image_name(cls, time_unit) -> str:
        return f"docs_by_{time_unit}_and_lang.png"

    @classmethod
    def get_chart_image_path(cls, time_unit) -> str:
        # E.g. ../lk_acts_data/data/lk_acts/docs_by_year_and_lang.png
        return os.path.join(
            cls.get_dir_docs_for_cls(), cls.get_chart_image_name(time_unit)
        )

    @classmethod
    def get_raw_remote_chart_image_url(cls, time_unit) -> str:
        # E.g. https://raw.githubusercontent.com/nuuuwan/lk_appeal_court_judgements/refs/heads/data/data/lk_appeal_court_judgements/docs_by_year_and_lang.png # noqa: E501
        return "/".join(
            [
                cls.get_raw_remote_data_branch_url(),
                cls.get_dir_docs_for_cls_relative(),
                cls.get_chart_image_name(time_unit),
            ]
        )

    @classmethod
    def __prepare_chart_data__(cls, ts_to_lang_n: dict):
        ts = sorted(ts_to_lang_n.keys())
        langs = sorted(
            {lang for v in ts_to_lang_n.values() for lang in v.keys()}
        )
        counts = {
            lang: [ts_to_lang_n.get(year, {}).get(lang, 0) for year in ts]
            for lang in langs
        }
        return ts, langs, counts

    @classmethod
    def __plot_stacked_bars__(cls, ax, ts_list, counts):
        bottom = np.zeros(len(ts_list))
        for lang in cls.LANGS:
            values = counts.get(lang, [])
            if values:
                color = cls.COLOR_MAP.get(lang, "grey")
                ax.bar(
                    ts_list, values, bottom=bottom, label=lang, color=color
                )
                bottom += np.array(values)
        return bottom

    @classmethod
    def __compute_xticks__(cls, ts_list: list[int]) -> list[int]:
        if len(ts_list) <= cls.MAX_X_LABELS:
            return ts_list
        step = max(1, len(ts_list) // cls.MAX_X_LABELS)
        xticks = ts_list[::step]
        if xticks[-1] != ts_list[-1]:
            xticks.append(ts_list[-1])
        return xticks

    @classmethod
    def __configure_axes__(cls, ax, ts_list, time_unit):
        x_label = time_unit.title()
        ax.set_xlabel(x_label)
        ax.set_ylabel("Number of documents")
        ax.set_title(
            f"Number of {cls.get_doc_class_label()} by {x_label} & Language"
        )
        xticks = cls.__compute_xticks__(ts_list)
        ax.set_xticks(xticks)
        ax.set_xticklabels([str(y) for y in xticks], rotation=45)
        ax.legend(title="Language")
        ax.yaxis.set_major_locator(MaxNLocator(integer=True))
        plt.tight_layout()

    @classmethod
    def __save_chart__(cls, fig, time_unit):
        image_path = cls.get_chart_image_path(time_unit)
        os.makedirs(os.path.dirname(image_path), exist_ok=True)
        fig.savefig(image_path, dpi=300)
        plt.close(fig)
        log.info(f"Wrote {image_path}")

    @classmethod
    def build_chart_by_time_and_lang(cls, ts_to_lang_n: dict, time_unit):
        ts_list, _, counts = cls.__prepare_chart_data__(ts_to_lang_n)
        fig, ax = plt.subplots(figsize=cls.FIG_SIZE)
        cls.__plot_stacked_bars__(ax, ts_list, counts)
        cls.__configure_axes__(ax, ts_list, time_unit)
        cls.__save_chart__(fig, time_unit)



import os

from utils import FileOrDirectory, JSONFile, Log, Time, TimeFormat

log = Log("AbstractDocSummaryMixin")


class AbstractDocSummaryMixin:

    @property
    def has_pdf(self) -> bool:
        return False

    @property
    def has_excel(self) -> bool:
        return False

    @property
    def has_tabular(self) -> bool:
        return False

    @staticmethod
    def get_url_source_list(doc_list) -> list:
        url_metadata_list = [doc.url_metadata for doc in doc_list]
        top_url_list = [
            "/".join(url.split("/")[:3]) for url in url_metadata_list
        ]
        top_url_to_n = {}
        for top_url in top_url_list:
            if top_url not in top_url_to_n:
                top_url_to_n[top_url] = 0
            top_url_to_n[top_url] += 1

        sorted_top_url_to_n = dict(
            sorted(
                top_url_to_n.items(), key=lambda item: item[1], reverse=True
            )
        )
        return list(sorted_top_url_to_n.keys())

    @classmethod
    def get_summary(cls) -> dict:
        doc_list = cls.list_all()
        doc_class_label = cls.get_doc_class_label()
        doc_class_emoji = cls.get_doc_class_emoji()
        doc_class_description = cls.get_doc_class_description()
        time_updated = TimeFormat.TIME.format(Time.now())
        n_docs = len(doc_list)
        n_docs_with_pdfs = len([doc for doc in doc_list if doc.has_pdf])
        n_docs_with_text = len([doc for doc in doc_list if doc.has_text])
        n_docs_with_excel = len([doc for doc in doc_list if doc.has_excel])
        n_docs_with_tabular = len(
            [doc for doc in doc_list if doc.has_tabular]
        )
        date_strs = [doc.date_str for doc in doc_list]
        date_str_min = min(date_strs)
        date_str_max = max(date_strs)
        dataset_size = FileOrDirectory(cls.get_data_branch_dir_root()).size
        latest_doc = doc_list[0]
        url_source_list = cls.get_url_source_list(doc_list)
        url_data = cls.get_remote_data_url_for_class()
        latest_doc_d = latest_doc.to_dict()
        langs = set([doc.lang for doc in doc_list])

        time_unit = cls.get_best_time_unit()
        ts_to_lang_to_n = cls.get_ts_to_lang_to_n(time_unit)
        cls.build_chart_by_time_and_lang(ts_to_lang_to_n, time_unit)
        url_chart = cls.get_raw_remote_chart_image_url(time_unit)

        return dict(
            doc_class_label=doc_class_label,
            doc_class_emoji=doc_class_emoji,
            doc_class_description=doc_class_description,
            time_updated=time_updated,
            n_docs=n_docs,
            n_docs_with_pdfs=n_docs_with_pdfs,
            n_docs_with_text=n_docs_with_text,
            n_docs_with_excel=n_docs_with_excel,
            n_docs_with_tabular=n_docs_with_tabular,
            date_str_min=date_str_min,
            date_str_max=date_str_max,
            dataset_size=dataset_size,
            url_source_list=url_source_list,
            url_data=url_data,
            url_chart=url_chart,
            langs=list(langs),
            latest_doc_d=latest_doc_d,
        )

    @classmethod
    def get_summary_path(cls) -> str:
        # E.g. ../lk_acts_data/data/lk_acts/summary.json
        return os.path.join(cls.get_dir_docs_for_cls(), "summary.json")

    @classmethod
    def build_summary(cls):
        summary = cls.get_summary()
        log.debug(f"{summary=}")
        summary_json_path = cls.get_summary_path()
        os.makedirs(cls.get_main_branch_dir_root(), exist_ok=True)
        JSONFile(summary_json_path).write(summary)
        log.info(f"Wrote {summary_json_path}")



from functools import cached_property

from utils import File, FileOrDirectory, Log

from scraper.abstract_doc.readme.AbstractDocReadMeMixin import \
    AbstractDocReadMeMixin
from scraper.global_readme.GlobalReadMeSummaryMixin import \
    GlobalReadMeSummaryMixin

log = Log("GlobalReadMe")


class GlobalReadMe(GlobalReadMeSummaryMixin):
    PATH = "README.md"

    def __init__(self, repo_to_doc_classes: dict[str, list[str]]):
        self.repo_to_doc_classes = repo_to_doc_classes
        self.summary_list = self.get_summary_list()
        self.global_summary = self.get_global_summary(self.summary_list)

    @cached_property
    def lines_for_global_summary(self) -> list[str]:
        log.debug(f"global_summary={self.global_summary}")
        all_dataset_size_humanized = FileOrDirectory.humanize_size(
            self.global_summary["all_dataset_size"]
        )
        return [
            f"**{self.global_summary['n_datasets']:,}** datasets, "
            + f"with **{self.global_summary['n_docs']:,}** documents"
            + f" (**{all_dataset_size_humanized}**).",
            "",
        ]

    def get_lines_for_dataset(self, i_dataset, summary) -> list[str]:
        header_lines = AbstractDocReadMeMixin.get_lines_for_header(summary)
        first_header_line = header_lines[0]
        first_header_line = f"## {i_dataset:03d} " + (
            first_header_line.replace("#SriLanka 🇱🇰", "")
            .replace("`Dataset`", "")
            .strip()[2:]
        )
        header_lines[0] = first_header_line
        return (
            header_lines
            + AbstractDocReadMeMixin.get_lines_for_blurb(summary)
            + AbstractDocReadMeMixin.get_lines_chart_docs_by_year_and_lang(
                summary
            )
        )

    @cached_property
    def lines_for_datasets(self) -> list[str]:
        lines = []
        for i_dataset, summary in enumerate(self.summary_list, start=1):
            lines.extend(self.get_lines_for_dataset(i_dataset, summary))
        return lines

    @cached_property
    def lines_for_footer(self) -> list[str]:
        return AbstractDocReadMeMixin.get_lines_for_footer()

    @cached_property
    def lines(self) -> list[str]:
        return (
            [
                "# 🇱🇰 #SriLanka `Datasets`",
                "",
            ]
            + self.lines_for_global_summary
            + self.lines_for_datasets
            + ["---", ""]
            + self.lines_for_footer
        )

    def build(self) -> None:
        File(self.PATH).write_lines(self.lines)
        log.info(f"Wrote {self.PATH}")



import os
import shutil

from utils import Log

log = Log("AbstractDocPipelineCleanupMixin")


class AbstractDocPipelineCleanupMixin:

    @classmethod
    def cleanup_doc_old_tabular_dirs(cls, doc):
        for dir_name in ["tabular", "worksheets"]:
            dir_path = os.path.join(doc.dir_doc, dir_name)
            if os.path.exists(dir_path):
                log.info(f"❌ Removing old dir: {dir_path}")
                shutil.rmtree(dir_path)

    @classmethod
    def cleanup_doc(cls, doc):
        cls.cleanup_doc_old_tabular_dirs(doc)

    @classmethod
    def cleanup_all(cls):
        docs = cls.list_all()
        for doc in docs:
            cls.cleanup_doc(doc)



from scraper.abstract_doc.core.AbstractDocBase import AbstractDocBase
from scraper.abstract_doc.core.AbstractDocBasePathsMixin import \
    AbstractDocBasePathsMixin
from scraper.abstract_doc.core.AbstractDocGeneratorMixin import \
    AbstractDocGeneratorMixin
from scraper.abstract_doc.core.AbstractDocMetadataMixin import \
    AbstractDocMetadataMixin
from scraper.abstract_doc.core.AbstractDocRemotePathMixin import \
    AbstractDocRemotePathMixin
from scraper.abstract_doc.core.AbstractDocTextMixin import AbstractDocTextMixin
from scraper.abstract_doc.hugging_face import AbstractDocHuggingFaceMixin
from scraper.abstract_doc.pipeline.AbstractDocPipelineMixin import \
    AbstractDocPipelineMixin
from scraper.abstract_doc.readme import AbstractDocReadMeMixin
from scraper.abstract_doc.readme.AbstractDocChartDocsByTimeAndLangMixin import \
    AbstractDocChartDocsByTimeAndLangMixin
from scraper.abstract_doc.readme.AbstractDocSummaryMixin import \
    AbstractDocSummaryMixin


class AbstractDoc(
    AbstractDocBase,
    AbstractDocBasePathsMixin,
    AbstractDocGeneratorMixin,
    AbstractDocMetadataMixin,
    AbstractDocRemotePathMixin,
    AbstractDocReadMeMixin,
    AbstractDocSummaryMixin,
    AbstractDocHuggingFaceMixin,
    AbstractDocPipelineMixin,
    AbstractDocChartDocsByTimeAndLangMixin,
    AbstractDocTextMixin,
):
    pass



import os
from functools import cached_property

from utils import File, Log

log = Log("AbstractDocTextMixin")


class AbstractDocTextMixin:

    @cached_property
    def text_path(self) -> str:
        return os.path.join(self.dir_doc, "doc.txt")

    @property
    def has_text(self) -> bool:
        return os.path.exists(self.text_path)

    def get_text(self):
        if not self.has_text:
            return None
        return File(self.text_path).read()



import os
from dataclasses import dataclass
from functools import cached_property

import pandas as pd
from utils import WWW, Log

from scraper.abstract_doc.AbstractDoc import AbstractDoc
from scraper.abstract_doc.data_mixins.AbstractTabularMixin import \
    AbstractTabularMixin

log = Log("AbstractExcelSpreadsheet")


@dataclass
class AbstractExcelSpreadsheet(AbstractTabularMixin, AbstractDoc):
    url_excel: str

    @cached_property
    def excel_path(self) -> str:
        return os.path.join(self.dir_doc, "doc.xlsx")

    @property
    def has_excel(self) -> bool:
        return os.path.exists(self.excel_path)

    def download_excel(self):
        try:
            WWW(self.url_excel).download_binary(self.excel_path)
        except Exception as e:
            log.error(f"Failed to download Excel from {self.url_excel}: {e}")
            return

    # ----------------------------------------------------------------
    # Worksheets (extracted from Excel)
    # ----------------------------------------------------------------

    @staticmethod
    def open_excel(excel_path: str) -> pd.ExcelFile:
        engine = "openpyxl" if excel_path.endswith(".xlsx") else "xlrd"
        return pd.ExcelFile(excel_path, engine=engine)

    def extract_tabular(self):
        if not os.path.exists(self.excel_path):
            return
        try:
            excel = self.open_excel(self.excel_path)
        except Exception as e:
            log.error(f"Failed to open Excel file {self.excel_path}: {e}")
            return

        for i_sheet, sheet_name in enumerate(excel.sheet_names, 1):
            sheet_name_cleaned = sheet_name.replace("/", "_").replace(
                " ", "_"
            )
            csv_path = os.path.join(
                self.dir_tabular,
                f"{i_sheet:02d}-{sheet_name_cleaned}.csv",
            )
            if os.path.exists(csv_path):
                continue

            df = excel.parse(sheet_name)
            os.makedirs(self.dir_tabular, exist_ok=True)
            df.to_csv(csv_path, index=False)
            log.info(f"Wrote {csv_path}")

    # ----------------------------------------------------------------
    # Scrape (ALL)
    # ----------------------------------------------------------------
    def scrape_extended_data_for_doc(self):
        if not self.has_excel:
            self.download_excel()

        self.scrape_extended_data_for_tabular_mixin()



import os

import pandas as pd
from datasets import Dataset
from utils import BigJSONFile, Chunker, Hash, Log

log = Log("AbstractDocHuggingFaceMixin")


class AbstractDocHuggingFaceMixin:
    MAX_CHUNK_SIZE = 2000
    MIN_OVERLAP_SIZE = 200
    HUGGING_FACE_USERNAME = os.environ.get("HUGGING_FACE_USERNAME")
    HUGGING_FACE_TOKEN = os.environ.get("HUGGING_FACE_TOKEN")

    @classmethod
    def get_dir_hugging_face_data(cls):
        # E.g. # e.g. ../lk_acts_data/data/lk_acts/hugging_face_data
        return os.path.join(cls.get_dir_docs_for_cls(), "hugging_face_data")

    @classmethod
    def get_docs_json_path(cls):
        return os.path.join(cls.get_dir_hugging_face_data(), "docs")

    @classmethod
    def build_docs(cls):
        d_list = [doc.to_dict() for doc in cls.list_all_ascending()]
        BigJSONFile(cls.get_docs_json_path()).write(d_list)
        return d_list

    @classmethod
    def get_data_list_for_doc(cls, doc):
        chunks = Chunker.chunk(
            doc.get_text(),
            cls.MAX_CHUNK_SIZE,
            cls.MIN_OVERLAP_SIZE,
        )
        d_list = []
        for chunk_index, chunk_text in enumerate(chunks):
            chunk_id = f"{doc.doc_id}-{chunk_index:04d}"
            d = doc.to_dict() | dict(
                chunk_id=chunk_id,
                chunk_index=chunk_index,
                language=doc.lang,
                md5=Hash.md5(chunk_text),
                chunk_size_bytes=len(chunk_text.encode("utf-8")),
                chunk_text=chunk_text,
            )
            d_list.append(d)

        return d_list

    @classmethod
    def get_chunks_json_path(cls):
        return os.path.join(cls.get_dir_hugging_face_data(), "chunks")

    @classmethod
    def build_chunks(cls):
        d_list = []
        for doc in cls.list_all_ascending():
            d_list.extend(cls.get_data_list_for_doc(doc))

        BigJSONFile(cls.get_chunks_json_path()).write(d_list)
        return d_list

    @classmethod
    def get_hugging_face_project(cls):
        label = cls.get_doc_class_label().replace("_", "-").lower()
        return f"{cls.HUGGING_FACE_USERNAME}/{label}"

    @classmethod
    def get_dataset_id(cls, label_suffix: str) -> str:
        return f"{cls.get_hugging_face_project()}-{label_suffix}"

    @classmethod
    def get_dataset_url(cls, label_suffix: str) -> str:
        return (
            "https://huggingface.co/datasets"
            + f"/{cls.get_dataset_id(label_suffix)}"
        )

    @classmethod
    def upload_to_hugging_face(cls):
        docs_df = pd.DataFrame(cls.build_docs())
        chunks_df = pd.DataFrame(cls.build_chunks())
        docs_ds = Dataset.from_pandas(docs_df)
        chunks_ds = Dataset.from_pandas(chunks_df)
        assert cls.HUGGING_FACE_USERNAME
        assert cls.HUGGING_FACE_TOKEN

        for ds, suffix in [(docs_ds, "docs"), (chunks_ds, "chunks")]:
            dataset_id = cls.get_dataset_id(suffix)
            repo_id = ds.push_to_hub(dataset_id, token=cls.HUGGING_FACE_TOKEN)
            log.info(f"🤗 Uploaded {dataset_id} to {repo_id}")

    @classmethod
    def build_and_upload_to_hugging_face(cls):
        try:
            cls.upload_to_hugging_face()
        except Exception as e:
            log.error(f"Error uploading to Hugging Face: {e}")



import os
import shutil
import tempfile
from dataclasses import dataclass
from functools import cached_property

import camelot
from utils import WWW, File, JSONFile, Log, PDFFile

from scraper.abstract_doc import AbstractDoc
from scraper.abstract_doc.data_mixins.AbstractTabularMixin import \
    AbstractTabularMixin

log = Log("AbstractPDFDoc")


@dataclass
class AbstractPDFDoc(AbstractTabularMixin, AbstractDoc):
    url_pdf: str

    # ----------------------------------------------------------------
    # PDF
    # ----------------------------------------------------------------
    @cached_property
    def pdf_path(self) -> str:
        return os.path.join(self.dir_doc, "doc.pdf")

    @property
    def has_pdf(self) -> bool:
        return os.path.exists(self.pdf_path)

    def download_pdf(self):
        with tempfile.NamedTemporaryFile(
            suffix=".pdf", delete=False
        ) as temp_pdf_f:
            temp_pdf_path = temp_pdf_f.name

            try:
                WWW(self.url_pdf).download_binary(temp_pdf_path)
            except Exception as e:
                log.error(f"Failed to download PDF from {self.url_pdf}: {e}")
                return

            try:
                PDFFile(temp_pdf_path).compress(self.pdf_path)
            except Exception as e:
                log.warning(
                    f"Failed to compress PDF from {self.url_pdf}: {e}"
                )
                shutil.move(temp_pdf_path, self.pdf_path)

    # ----------------------------------------------------------------
    # Blocks (Extracted from PDF)
    # ----------------------------------------------------------------
    @cached_property
    def blocks_path(self) -> str:
        return os.path.join(self.dir_doc, "blocks.json")

    @property
    def has_blocks(self) -> bool:
        return os.path.exists(self.blocks_path)

    def extract_blocks(self):
        if not self.has_pdf:
            return
        pdf_file = PDFFile(self.pdf_path)
        blocks = pdf_file.get_blocks()
        JSONFile(self.blocks_path).write(blocks)
        log.info(f"Wrote {self.blocks_path} ({len(blocks):,} blocks)")

    def get_blocks(self):
        return JSONFile(self.blocks_path).read()

    # ----------------------------------------------------------------
    # Worksheets (extracted from PDF)
    # ----------------------------------------------------------------

    @staticmethod
    def get_tables(pdf_path: str) -> camelot.core.TableList | None:
        try:
            tables = camelot.read_pdf(
                pdf_path,
                flavor="lattice",
                pages="all",
                process_background=True,
                line_scale=60,
            )
            n_tables = len(tables)
            log.debug(f"Found {n_tables} tables in {pdf_path}")
            return tables
        except Exception as e:
            log.error(f"Failed to extract tables from {pdf_path}: {e}")
            return None

    def extract_tabular(self):
        if not self.has_pdf:
            return

        tables = self.get_tables(self.pdf_path)
        if not tables or len(tables) == 0:
            log.info(f"No tables found in {self.pdf_path}")
            return

        os.makedirs(self.dir_tabular, exist_ok=True)
        tables.export(os.path.join(self.dir_tabular, "table.csv"), f="csv")

    # ----------------------------------------------------------------
    # Text (From Blocks)
    # ----------------------------------------------------------------

    def get_text_from_block(self):
        if not self.has_pdf:
            return None
        blocks = self.get_blocks()
        text_list = [block["text"] for block in blocks if block["text"]]
        content = "\n\n".join(text_list)
        return content

    def get_text_all(self):
        lines = []
        for text, label in [
            (self.get_text_from_block(), "BLOCKS"),
            (self.get_text_from_tabular(), "WORKSHEETS"),
        ]:
            if text:
                lines.append(f"==== {label} ==== ")
                lines.append("")
                lines.append(text)
                lines.append("")

        return "\n".join(lines).strip()

    def extract_text(self):
        if not self.has_pdf:
            return None
        File(self.text_path).write(self.get_text_all())
        log.info(f"Wrote {self.text_path}")

    # ----------------------------------------------------------------
    # All
    # ----------------------------------------------------------------
    def scrape_extended_data_for_doc(self):
        if not self.has_pdf:
            self.download_pdf()

        if self.has_pdf and not self.has_blocks:
            self.extract_blocks()

        if self.has_pdf and not self.has_tabular:
            self.extract_tabular()

        if self.has_blocks and not self.has_text:
            self.extract_text()



# scraper.abstract_doc.readme (auto generate by build_inits.py)
# flake8: noqa: F408

from scraper.abstract_doc.readme.AbstractDocChartDocsByTimeAndLangMixin import \
    AbstractDocChartDocsByTimeAndLangMixin
from scraper.abstract_doc.readme.AbstractDocReadMeMixin import \
    AbstractDocReadMeMixin
from scraper.abstract_doc.readme.AbstractDocSummaryMixin import \
    AbstractDocSummaryMixin



import sys

from utils import Log

from scraper.abstract_doc.pipeline.AbstractDocPipelineCleanupMixin import \
    AbstractDocPipelineCleanupMixin  # noqa: E501
from scraper.abstract_doc.pipeline.AbstractDocPipelineExtendedDataMixin import \
    AbstractDocPipelineExtendedDataMixin  # noqa: E501
from scraper.abstract_doc.pipeline.AbstractDocPipelineMetadataMixin import \
    AbstractDocPipelineMetadataMixin  # noqa: E501
from scraper.global_readme.GlobalReadMe import GlobalReadMe

log = Log("AbstractDocPipelineMixin")


class AbstractDocPipelineMixin(
    AbstractDocPipelineMetadataMixin,
    AbstractDocPipelineExtendedDataMixin,
    AbstractDocPipelineCleanupMixin,
):
    MAX_DT = 750.0

    @classmethod
    def run_pipeline(cls, max_dt=None):
        max_dt = (
            max_dt
            or (float(sys.argv[1]) if len(sys.argv) > 1 else None)
            or AbstractDocPipelineMixin.MAX_DT
        )
        log.debug(f"{max_dt=}s")
        cls.cleanup_all()
        cls.scrape_all_metadata(max_dt)
        cls.write_all()
        cls.scrape_all_extended_data(max_dt)
        cls.build_summary()
        cls.build_doc_class_readme()
        cls.build_and_upload_to_hugging_face()

        if not cls.is_multi_doc():
            GlobalReadMe(
                {cls.get_repo_name(): [cls.get_doc_class_label()]}
            ).build()



import os

from utils import File, Log

log = Log("AbstractTabularMixin")


class AbstractTabularMixin:

    @property
    def dir_tabular(self) -> str:
        return os.path.join(self.dir_doc, "tabular_data")

    @property
    def has_tabular(self) -> bool:
        return os.path.exists(self.dir_tabular) and bool(
            os.listdir(self.dir_tabular)
        )

    def extract_tabular(self) -> None:
        raise NotImplementedError

    # ----------------------------------------------------------------
    # Text (extracted from Worksheet CSVs)
    # ----------------------------------------------------------------

    def get_text_from_tabular(self):
        if os.path.exists(self.text_path):
            return None
        if not os.path.exists(self.dir_tabular):
            return None

        content_groups = []
        for file_name in sorted(os.listdir(self.dir_tabular)):
            if not file_name.endswith(".csv"):
                continue
            csv_path = os.path.join(self.dir_tabular, file_name)
            content = File(csv_path).read()
            content_groups.append(f"--- {file_name} ---\n")
            content_groups.append(content)
        return "\n".join(content_groups)

    def extract_text_for_tabular_mixin(self) -> None:
        content = self.get_text_from_tabular()
        if not content:
            return
        text_file = File(self.text_path)
        text_file.write(content)
        log.info(f"Wrote {text_file}")

    def extract_text(self) -> None:
        return self.extract_text_for_tabular_mixin()

    # ----------------------------------------------------------------
    # Scrape (ALL)
    # ----------------------------------------------------------------
    def scrape_extended_data_for_tabular_mixin(self):
        if not os.path.exists(self.dir_tabular):
            self.extract_tabular()

        if not self.has_text:
            self.extract_text_for_tabular_mixin()



# scraper.abstract_doc.hugging_face (auto generate by build_inits.py)
# flake8: noqa: F408

from scraper.abstract_doc.hugging_face.AbstractDocHuggingFaceMixin import \
    AbstractDocHuggingFaceMixin
